package be.iminds.ilabt.jfed.ui.cli2.wizard;

import be.iminds.ilabt.jfed.experiment.setup.ExperimentSetupConfig;
import be.iminds.ilabt.jfed.rspec.adv_based_generator.AdvertisementBasedRspecGeneratorConfig;
import be.iminds.ilabt.jfed.ui.cli2.instruction.*;
import com.fasterxml.jackson.annotation.JsonProperty;
import org.jetbrains.annotations.Contract;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.*;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static be.iminds.ilabt.jfed.experiment.setup.ExperimentSetupConfig.RequestRSpecSource.ALREADY_PROVISIONED;

public class ActionWizard {
    private final PrintStream out;
    private final PrintStream err;
    private final InputStream in;
    private final BufferedReader readin;

    //filled in by askGenericQuestions()
    private User user;
    private List<ActionOutput> actionOutputs;
    private List<DebugOutput> debugOutputs;
    private List<CallOutput> callOutputs;
    private Instruction.ProxySettings proxy;

    //filled in by askExistingSliceQuestions() and askNewSliceQuestions()
    private ExperimentSetupConfig.Slice slice;

    public ActionWizard(@Nonnull PrintStream out,
                        @Nonnull PrintStream err,
                        @Nonnull InputStream in) {
        this.out = out;
        this.err = err;
        this.in = in;
        this.readin = new BufferedReader(new InputStreamReader(in));
    }

    @Contract("_,!null,_,_,_ -> !null; _,_,_,false,_ -> !null")
    private <T> T ask(String question, T defaultValue, Function<String, T> parser, boolean allowNull, String defaultShownOverride) throws IOException {
        while (true) {
            out.print(question + "? ");
            if (defaultValue != null && defaultShownOverride == null)
                out.print("(default=" + defaultValue + ") ");
            if (defaultShownOverride != null)
                out.print("(" + defaultShownOverride + ") ");
            String val = readin.readLine();
            out.println();
            if (val.trim().isEmpty() && defaultValue != null) return defaultValue;
            T res = parser.apply(val);
            if (res != null)
                return res;
            else {
                if (allowNull)
                    return null;
                else
                    err.println("Illegal answer, try again.");
            }
        }
    }
    @Contract("_,!null,_,_,_ -> !null; _,_,_,false,_ -> !null")
    private <T> T askMultiLine(String question, T defaultValue, Function<String, T> parser, boolean allowNull, String defaultShownOverride) throws IOException {
        while (true) {
            out.print(question + "? ");
            if (defaultValue != null && defaultShownOverride == null)
                out.print("(default=" + defaultValue + ") ");
            if (defaultShownOverride != null)
                out.print("(" + defaultShownOverride + ") ");
            StringBuilder val = new StringBuilder();
            String line =readin.readLine();
            val.append(line+"\n");
            while (!line.trim().equals("EOF")) {
                val.append(line+"\n");
                line = readin.readLine();
            }
            out.println();
            String fval = val.toString();
            if (fval.trim().isEmpty() && defaultValue != null) return defaultValue;
            T res = parser.apply(fval);
            if (res != null)
                return res;
            else {
                if (allowNull)
                    return null;
                else
                    err.println("Illegal answer, try again.");
            }
        }
    }
    public String askString(String question, String defaultValue, boolean allowNull) throws IOException {
        return ask(question, defaultValue, s -> s.trim().isEmpty() ? null : s, allowNull, null);
    }
    public String askMultiString(String question, String defaultValue, boolean allowNull) throws IOException {
        return askMultiLine(question, defaultValue, s -> s.trim().isEmpty() ? null : s, allowNull, null);
    }
    public int askInt(String question, Integer defaultValue) throws IOException {
        return ask(question, defaultValue,
                (s) -> {
                    try {
                        return Integer.parseInt(s);
                    } catch (NumberFormatException e) {
                        return null;
                    }
                },
                false, null);
    }
    public long askLong(String question, Long defaultValue) throws IOException {
        return ask(question, defaultValue,
                (s) -> {
                    try {
                        return Long.parseLong(s);
                    } catch (NumberFormatException e) {
                        return null;
                    }
                },
                false, null);
    }
    public Long askLongOrNull(String question, Long defaultValue) throws IOException {
        return ask(question, defaultValue,
                (s) -> {
                    try {
                        return Long.parseLong(s);
                    } catch (NumberFormatException e) {
                        return null;
                    }
                },
                true, null);
    }
    public boolean askBool(String question, Boolean defaultValue) throws IOException {
        return ask(question, defaultValue,
                (s) -> {
                    if (s == null) return null;
                    if (Arrays.asList("true", "yes", "t", "y", "1").contains(s.toLowerCase()))
                        return true;
                    if (Arrays.asList("false", "no", "f", "n", "0").contains(s.toLowerCase()))
                        return false;
                    return null;
                },
                false,
                defaultValue == null ? "y/n" : (defaultValue ? "Y/n" : "y/N"));
    }

    public List<String> askStringList(@Nonnull String question, boolean allowEmpty) throws IOException {
        String csvList = askString(question, null, allowEmpty);
        if (csvList == null || csvList.trim().isEmpty()) {
            return Collections.emptyList();
        } else {
            return Arrays.asList(csvList.trim().split(","));
        }
    }

    public int askListPosition(@Nonnull String question, @Nonnull List<String> options, @Nullable Integer defaultValue) throws IOException {
        assert options.size() > 0;
        for (int i = 0; i < options.size(); i++) {
            out.println(i+") "+options.get(i));
        }
        out.println();
        return ask(question, defaultValue,
                (s) -> {
                    try {
                        int res = Integer.parseInt(s);
                        if (res < 0 || res > options.size())
                            return null;
                        return res;
                    } catch (NumberFormatException e) {
                        return null;
                    }
                },
                false, null);
    }
    public <T extends Enum> T askEnum(@Nonnull String question, @Nonnull Class<T> enumClass, @Nullable T defaultValue) throws IOException {
        T[] values = enumClass.getEnumConstants();
        for (int i = 0; i < values.length; i++) {
            assert values[i].ordinal() == i;
            out.println(i+") "+values[i].name());
        }
        out.println();
        return ask(question, defaultValue,
                (s) -> {
                    try {
                        int res = Integer.parseInt(s);
                        if (res < 0 || res > values.length)
                            return null;
                        return values[res];
                    } catch (NumberFormatException e) {
                        return null;
                    }
                },
                false, null);
    }
    public String askListValue(@Nonnull String question, @Nonnull List<String> options, @Nullable Integer defaultValue) throws IOException {
        int res = askListPosition(question, options, defaultValue);
        return options.get(res);
    }
    public boolean askBoolChoice(@Nonnull String trueCase, @Nonnull String falseCase, Boolean defaultValue) throws IOException {
        int res = askListPosition("Choose", Arrays.asList(trueCase, falseCase), defaultValue == null ? null : (defaultValue ? 0 : 1));
        return res == 0;
    }

    @Nullable
    public Instruction generate() throws IOException {
        out.println("Welcome to the jFed CLI Wizard. This tool guide you through all the available actions and options.");
        out.println("This will only create the YaML config for the action, it will not execute the action.");
        out.println("At the end, it will be shown how the created action can be executed.");
        out.println();
        out.println("Available actions:");
        out.println();

        for (Instruction.Action action : Instruction.Action.values()) {
            out.println(action.ordinal()+") "+action.getNiceName());
        }
        out.println();
        int actionOrdinal = askInt("Action", null);

        Instruction.Action action = Instruction.Action.values()[actionOrdinal];
        out.println("You chose action: "+action.name());

        askGenericQuestions();

        switch (action) {
            case USERINFO: return generateUserInfo();
            case SLICEINFO: return generateSliceInfo();
            case SLICEACTION: return generateSliceAction();
            case RUN: return generateRun();
            case CREATESLICE: return generateCreateSlice();
            case LISTRESOURCES: return generateListResources();
            case MANIFEST: return generateManifest();
            case DELETE: return generateDelete();
            case POA: return generatePoa();
            case RENEW: return generateRenew();
            case SLICECREDENTIAL: return generateSliceCredential();
            case LOGIN: return generateLogin();
            case DECRYPTPEM: return generateDecryptPem();
            default: throw new RuntimeException("Action "+action.name()+" not yet supported in wizard.");
        }
    }

    public void askGenericQuestions() throws IOException {
        out.println("\n*** Generic questions:\n");
        askUserQuestions();
        askProxyQuestions();
        askOutputQuestions();
    }
    /*
    *  List<String> pem,
                    @JsonProperty("passwordMethod") @Nullable PasswordMethod passwordMethod,
                    @JsonProperty("password") @Nullable String password,
                    @JsonProperty("speaksForCredential") @Nullable List<String> speaksForCredential
    * */
    public void askUserQuestions() throws IOException {
        String pem = askString("User PEM login filename", null, false);
        boolean hasPassword = askBool("Is this file password protected", true);
        if (!hasPassword) {
            user = new User(
                    Collections.singletonList(pem),
                    User.PasswordMethod.NONE,
                    null,
                    null);
            return;
        }

        if (askBool("Store the password in the generated action file", false)) {
            user = new User(
                    Collections.singletonList(pem),
                    User.PasswordMethod.DIRECT,
                    askString("Password", null, false),
                    null);
            return;
        }

        if (askBool("Store the password in a seperate file", false)) {
            user = new User(
                    Collections.singletonList(pem),
                    User.PasswordMethod.FILE,
                    askString("Password file", null, false),
                    null);
            return;
        }

        out.println("Password won't be stored. It will be asked each time you execute the action with the CLI.");
        user = new User(
                Collections.singletonList(pem),
                User.PasswordMethod.INTERACTIVE,
                null,
                null);
    }

    public void askProxyQuestions() throws IOException {
        if (askBool("Use same jFed proxy settings as in GUI", true)) {
            proxy = null;
            return;
        }

        proxy = new Instruction.ProxySettings(
                askBool("Use jFed proxy for API calls", false),
                askBool("Use jFed proxy for SSH", false)
        );
    }

    private class OutputPair {
        private final OutputTarget target;
        private final String filename;

        public OutputPair(OutputTarget target, String filename) {
            this.target = target;
            this.filename = filename;
        }

        public OutputTarget getTarget() {
            return target;
        }

        public String getFilename() {
            return filename;
        }
    }
    private OutputPair askOutputPair(String defaultVal) throws IOException {
        String out = askString("Where to write output? (STDOUT or STDERR or a filename) ", defaultVal, false);
        if (out.equalsIgnoreCase("stdout")) {
            return new OutputPair(OutputTarget.STDOUT, null);
        }
        if (out.equalsIgnoreCase("stderr")) {
            return new OutputPair(OutputTarget.STDERR, null);
        }
        return new OutputPair(OutputTarget.FILE, out);
    }

    public void askOutputQuestions() throws IOException {
        out.println("Most actions have in an \"action output\" (the requested information, a boolean indicating the operation success, ...)\n" +
                "You can choose in which format (text or json), and where (stream or file) this output should be written.\n" +
                "You can choose multiple outputs.\n" +
                "Default: text is sent to stdout.");
        if (askBool("Use default action output", true)) {
            actionOutputs = null;
        } else {
            actionOutputs = new ArrayList<>();
            if (askBool("Output action as plain text", true)) {
                OutputPair outputPair = askOutputPair("STDOUT");
                actionOutputs.add(new ActionOutput(
                        ActionOutput.Format.TXT, outputPair.target, outputPair.filename));
            }
            if (askBool("Output action as JSON", true)) {
                OutputPair outputPair = askOutputPair("STDOUT");
                actionOutputs.add(new ActionOutput(
                        ActionOutput.Format.JSON, outputPair.target, outputPair.filename));
            }
        }

        out.println("Debug output can be useful if things go wrong.\nDefault: no debug output.");
        if (!askBool("Use debug output", false)) {
            debugOutputs = null;
        } else {
            DebugOutput.Level level = askEnum("Debug Level", DebugOutput.Level.class, DebugOutput.Level.INFO);
            if (askBool("debug as full text", true)) {
                OutputPair outputPair = askOutputPair("debug.txt");
                debugOutputs = Collections.singletonList(new DebugOutput(level,
                        DebugOutput.Format.TXT_FULL, outputPair.target, outputPair.filename));
            } else {
                OutputPair outputPair = askOutputPair("STDOUT");
                debugOutputs = Collections.singletonList(new DebugOutput(level,
                        DebugOutput.Format.TXT, outputPair.target, outputPair.filename));
            }
        }

        out.println("API calls made by the CLI can be logged to a file. This can be very useful in debugging any issues.\n" +
                "Default: no API call logging.");
        if (!askBool("Write API calls to file", false)) {
            callOutputs = null;
        } else {
            String callFile = askString("Call filename", null, false);
            CallOutput.Format outputFormat = askEnum("Call Output Format", CallOutput.Format.class, CallOutput.Format.HTML);
            callOutputs = Collections.singletonList(new CallOutput(outputFormat, OutputTarget.FILE, callFile));
        }
    }


    private void askSliceQuestions(@Nullable Boolean failOnExistingSlice,
                                   @Nullable Boolean failOnNonExistingSlice,
                                   @Nullable Boolean renewExistingSliceIfNeeded,
                                   boolean askExpire) throws IOException {
        String sliceName = askString("Slice Name", null, false);


        Long expireTimeMin = askExpire ? (long) askInt("Expiration time in minutes", 120) : null;

        String project = askString("Project (leave empty for more options)", null, true);

        ExperimentSetupConfig.ProjectSource projectSource;
        boolean failIfNoProject;
        if (project != null && !project.trim().isEmpty()) {
            projectSource = ExperimentSetupConfig.ProjectSource.PROVIDED;
            failIfNoProject = true;
        } else {
            if (askBool("Determine project automatically", true)) {
                projectSource = ExperimentSetupConfig.ProjectSource.AUTO;
                failIfNoProject = true;
            } else {
                if (askBool("Do not use a project (not advised)", false)) {
                    projectSource = ExperimentSetupConfig.ProjectSource.NONE;
                    failIfNoProject = false;
                } else {
                    throw new RuntimeException("Out of options: You should probably just have provided a project name");
                }
            }
        }

        slice = new ExperimentSetupConfig.Slice(
                sliceName, failOnExistingSlice, failOnNonExistingSlice, expireTimeMin,
                renewExistingSliceIfNeeded, projectSource, failIfNoProject, project);
    }

    public void askExistingSliceQuestions() throws IOException {
        askExistingSliceQuestions(false);
    }
    public void askExistingSliceQuestions(boolean askExpire) throws IOException {
        out.println("\n*** Slice questions:\n");

        askSliceQuestions(false, true, false, askExpire);
    }

    public void askNewSliceQuestions() throws IOException {
        out.println("\n*** Slice questions:\n");

        askSliceQuestions(true, false, true, true);
    }

    public Instruction generateUserInfo() throws IOException {
        out.println("\n*** User Info request specific questions:\n");

        boolean showUserUrn = askBool("Show user URN", true);
        boolean listSlices = askBool("List slices", true);
        boolean listProjects = askBool("List projects", true);

        return new UserInfoInstruction(Instruction.Action.USERINFO, user, actionOutputs, debugOutputs, callOutputs, proxy,
                showUserUrn, listSlices, listProjects);
    }

    private Instruction generateSliceInfo() throws IOException {
        askExistingSliceQuestions();

        out.println("\n*** Slice Info request specific questions:\n");

        Boolean showStatus = askBool("Show Status", true);
        Boolean showMetaInfo = askBool("Show Meta Info", true);
        Boolean showUsers = askBool("Show Slice Users", true);
        Boolean showCredentials = askBool("Show Slice Credentials", false);
        return new SliceInfoInstruction(Instruction.Action.MANIFEST, user, actionOutputs, debugOutputs, callOutputs, proxy, slice,
                showStatus, showMetaInfo, showUsers, showCredentials);
    }

    private ExperimentSetupConfig.ESpec generateESpec() throws IOException {
        throw new RuntimeException("ESpec not yet supported in wizard.");
    }

    private Instruction generateSliceAction() throws IOException {
        askExistingSliceQuestions();

        out.println("\n*** Slice Action specific questions:\n");

        String sliceActionName = askListValue("Slice Action",
                Stream.of(SliceActionInstruction.SliceAction.values()).map(Enum::name).collect(Collectors.toList()),
                0);
        SliceActionInstruction.SliceAction sliceAction = SliceActionInstruction.SliceAction.valueOf(sliceActionName);

        ExperimentSetupConfig.ESpec espec = null;
        if (sliceAction == SliceActionInstruction.SliceAction.RerunESpec) {
            espec = generateESpec();
        }

        return new SliceActionInstruction(Instruction.Action.SLICEACTION, user, actionOutputs, debugOutputs, callOutputs, proxy,
                slice,
                sliceAction,
                espec);
    }

    private Instruction generateRun() throws IOException {
        out.println("\n*** Slice questions:\n");

        boolean failOnExistingSlice = askBool("Allow running in new slice", true);
        boolean failOnNonExistingSlice = askBool("Allow running in existing slice", false);
        boolean renewExistingSliceIfNeeded = failOnNonExistingSlice ? false :
                askBool("Renew existing slice if needed", true);

        askSliceQuestions(failOnExistingSlice, failOnNonExistingSlice, renewExistingSliceIfNeeded, true);

        out.println("\n*** Run questions:\n");

        // *** experiment ***
        ExperimentSetupConfig.RequestRSpec requestRSpec;
        int rspecEspecChoice = askListPosition("What do you want to use to define the experiment", Arrays.asList("Only Request RSpec", "Only ESpec", "(ADVANCED) Both RSpec and ESpec"), 0);
        if (rspecEspecChoice != 1) {
            ExperimentSetupConfig.RequestRSpecSource source =
                    askEnum("Request RSpec source",
                            ExperimentSetupConfig.RequestRSpecSource.class,
                            ExperimentSetupConfig.RequestRSpecSource.PROVIDE_FILE);
            Boolean autoAssignSpecificNodesUsingListResources = askBool("(ADVANCED) If the RSpec contains nodes without component_id, assign to available nodes using a ListResources call", false);
            Boolean useNitosLease = askBool("(ADVANCED) use Nitos lease (nitos testbed only)", false);

            switch (source) {
                case PROVIDE_URL: {
                    String providedContentSource = askString("URL of the Request RSpec", null, false);
                    String bindUnboundNodesUrn = askString("(ADVANCED) URL To bind unbound (= no component_manager_id) nodes to (leave empty to not bind them)", null, true);
                    requestRSpec = new ExperimentSetupConfig.ProvidedUrlRequestRSpec(
                            source,
                            providedContentSource,
                            false, //only makes sense for tests
                            bindUnboundNodesUrn,
                            autoAssignSpecificNodesUsingListResources, useNitosLease);
                    break;
                }
                case PROVIDE_FILE: {
                    String providedContentSource = askString("File of the Request RSpec", null, false);
                    String bindUnboundNodesUrn = askString("(ADVANCED) URL To bind unbound (= no component_manager_id) nodes to (leave empty to not bind them)", null, true);
                    requestRSpec = new ExperimentSetupConfig.ProvidedFileRequestRSpec(
                            source,
                            providedContentSource,
                            false, //only makes sense for tests
                            bindUnboundNodesUrn,
                            autoAssignSpecificNodesUsingListResources, useNitosLease);
                    break;
                }
                case PROVIDE_CONTENT: {
                    String providedContentSource = askMultiString("Request RSpec (use EOF on a seperate line to end input)", null, false);
                    String bindUnboundNodesUrn = askString("(ADVANCED) URL To bind unbound (= no component_manager_id) nodes to (leave empty to not bind them)", null, true);
                    requestRSpec = new ExperimentSetupConfig.ProvidedContentRequestRSpec(
                            source,  providedContentSource,
                            false, //only makes sense for tests
                            bindUnboundNodesUrn,
                            autoAssignSpecificNodesUsingListResources, useNitosLease);
                    break;
                }
                case GENERATE_SIMPLE: {
                    ExperimentSetupConfig.GenerateSimpleRequestRSpec defaultGen =
                            new ExperimentSetupConfig.GenerateSimpleRequestRSpec();
                    Integer nodeCount = askInt("Number of nodes in Request RSpec", defaultGen.getNodeCount());
                    String nodeSliverType = askString("Sliver Type of node(s)", defaultGen.getNodeSliverType(), true);
                    Boolean nodeExclusive = askBool("Is the \"exclusive\" attribute set for the nodes", defaultGen.getNodeExclusive());
                    List<String> nodeComponentUrn = askStringList("Component URNs of the nodes (comma separated list, may be empty)", true);
                    List<String> nodeAnsibleGroup = askStringList("Node ansible groups (comma separated list, may be empty)", true);
                    requestRSpec = new ExperimentSetupConfig.GenerateSimpleRequestRSpec(
                            source,
                            nodeCount,
                            nodeSliverType,
                            nodeExclusive,
                            nodeComponentUrn,
                            nodeAnsibleGroup,
                            autoAssignSpecificNodesUsingListResources, useNitosLease);
                    break;
                }
                case GENERATE_USING_JFED: {
                    Integer nodeCount = askInt("Number of nodes in Request RSpec", 1);
                    String jFedResourceClass = askString("\"ResourceClass\" to use (= jFed icon ID, see https://flsmonitor-api.fed4fire.eu/resourceclass for a full list)", null, true);
                    List<String> nodeAnsibleGroup = askStringList("Node ansible groups (comma separated list, may be empty)", true);
                    requestRSpec = new ExperimentSetupConfig.GenerateUsingJFedRequestRSpec(
                            source,
                            nodeCount,
                            jFedResourceClass,
                            nodeAnsibleGroup,
                            autoAssignSpecificNodesUsingListResources, useNitosLease);
                    break;
                }
                case GENERATE_USING_ADVERTISEMENT: {
                    AdvertisementBasedRspecGeneratorConfig config = new ExperimentSetupConfig.GenerateUsingAdvertisementRequestRSpec().getConfig();
                    out.println("(ADVANCED) The default AdvertisementBasedRspecGeneratorConfig will be used. Edit it in the result.");
                    requestRSpec = new ExperimentSetupConfig.GenerateUsingAdvertisementRequestRSpec(
                            source, autoAssignSpecificNodesUsingListResources, useNitosLease, config);
                    break;
                }
                case ALREADY_PROVISIONED: {
                    requestRSpec = new ExperimentSetupConfig.AlreadyProvisionedRequestRSpec(
                            source, autoAssignSpecificNodesUsingListResources, useNitosLease);
                    break;
                }
                default: throw new RuntimeException("Unsupported RSpec source: "+source);
            }

        } else {
            requestRSpec = null;
        }

        ExperimentSetupConfig.ESpec eSpec;
        if (rspecEspecChoice != 0) {
            ExperimentSetupConfig.ESpecSource source =
                    askEnum("ESpec source",
                            ExperimentSetupConfig.ESpecSource.class,
                            ExperimentSetupConfig.ESpecSource.PROVIDE_DIR);
            String providedContentSource;
            switch (source) {
                case PROVIDE_ARCHIVE_URL: {
                    providedContentSource = askString("ESpec Archive URL", null, false);
                    break;
                }
                case PROVIDE_ARCHIVE_FILE: {
                    providedContentSource = askString("ESpec Archive Filename", null, false);
                    break;
                }
                case PROVIDE_DIR: {
                    providedContentSource = askString("ESpec Directory", null, false);
                    break;
                }
                case PROVIDE_GIT_REPO_DIR: {
                    providedContentSource = askString("ESpec Git Repo Dir", null, false);
                    break;
                }
                case DIRECT: {
                    providedContentSource = askMultiString("ESpec experiment-specification.yml content (use EOF on a seperate line to end input)", null, false);
                    break;
                }
                default: throw new RuntimeException("Unsupported ESpec source: "+source);
            }
            List<ExperimentSetupConfig.ESpecLogStorageDetails> logStorage = null;
            //TODO: allow specifying ESpec log storage in wizard
            eSpec = new ExperimentSetupConfig.ESpec(source, providedContentSource, logStorage);
        } else {
            eSpec = null;
        }

        Boolean overrideESpecRSpec;
        if (eSpec != null && requestRSpec != null && requestRSpec.getSource() != ALREADY_PROVISIONED && rspecEspecChoice == 2) {
            out.println("You provided an RSpec, and an ESpec containing an RSpec. Which RSpec should be used?");
            overrideESpecRSpec = askBoolChoice("Use the RSpec provided seperatly",
                    "Use the RSpec inside of the ESpec",
                    true);
        } else {
            overrideESpecRSpec = null;
        }

        ExperimentSetupConfig.Provision provision; //Boolean useGeniEndTimeOptionForCreateSliver, Boolean enabled
        if (askBool("Do you want to start the experiment (=\"Provision the slice\")", true)) {
            provision = new ExperimentSetupConfig.Provision(
                    askBool("(ADVANCED) Do you want to use the geni_end_time option for CreateSliver", false),
                    true);
        } else {
            provision = new ExperimentSetupConfig.Provision(false, false);
        }
        ExperimentSetupConfig.WaitForReady waitForReady; //Long maxTimeMin, Long extraWaitTimeSeconds, Boolean enabled
        if (askBool("Do you want to wait until the experiment is ready", true)) {
            waitForReady = new ExperimentSetupConfig.WaitForReady(
                    askLongOrNull("What is the maximum wait time (in minutes) (leave empty to use a default)", null),
                    askLongOrNull("(ADVANCED) How long should jFed wait after the experiment is ready, before proceeding (this is only useful to work around some testbed specific bugs) (in minutes) (leave empty not to wait extra)", null),
                    false);
        } else {
            waitForReady = new ExperimentSetupConfig.WaitForReady(null, null, false);
        }
        Boolean runLinkTest = askBool("Do you want to run a link test", false);

        ExperimentSetupConfig experiment = new ExperimentSetupConfig(
                slice, requestRSpec, eSpec, provision, waitForReady, overrideESpecRSpec, runLinkTest);

        // *** shareWith ***
        RunInstruction.ShareWith shareWith;
        if (askBool("Share the experiment", false)) {
            boolean shareMembers = askBool("Share the experiment with all project members", false);

            List<String> shareUsernames = askStringList("Share the experiment with specific users (provide usernames, in a comma seperated list)", true);

            shareWith = new RunInstruction.ShareWith(shareUsernames, shareMembers);
        } else {
            shareWith = null;
        }

        // *** sshKeys ***
        RunInstruction.SshKeySource sshKeys;
        out.println("Questions about login SSH keys on requested nodes:");
        if (!askBool("   Use default login SSH keys", true)) {
            boolean userCert = askBool("   Add your login PEM SSH key", true);
            boolean userKeys = askBool("   Add the custom SSH key configured in jFed", true);
            boolean shareWithKeys = askBool("   Add SSH keys of all users the slice is shared with", true);
            boolean shareSshRspec = askBool("   Add all SSH keys specified in the RSpec (if any)", true);
            List<String> extraUsernames = askStringList("   Extra users of which to add SSH keys (provide usernames, in a comma seperated list)", true);
            List<String> extraKeys = askStringList("   Extra SSH keys (provide openssh format, in a comma seperated list)", true);
            sshKeys = new RunInstruction.SshKeySource(userCert, userKeys, shareWithKeys, shareSshRspec, extraUsernames, extraKeys);
        } else {
            sshKeys = null;
        }

        // *** deleteOn ***
        RunInstruction.DeleteOn deleteOn;
        if (!askBool("   Use default Experiment Delete conditions (= only delete on failed basic experiment creation)", true)) {
            boolean failCreate = askBool("Delete experiment if basic experiment creation fails", true);
            boolean failBecomeReady = askBool("Delete experiment if experiment fails to become ready", false);
            boolean failConnectivityTest = askBool("Delete experiment if connectivity test fails", false);
            boolean failLinkTest = askBool("Delete experiment if link test fails", false);
            boolean failESpec = askBool("Delete experiment if ESpec execution fails", false);
            deleteOn = new RunInstruction.DeleteOn(
                    failCreate, failBecomeReady, failConnectivityTest, failLinkTest, failESpec);
        } else {
            deleteOn = null;
        }

        return new RunInstruction(Instruction.Action.RUN, user, actionOutputs, debugOutputs, callOutputs, proxy,
                experiment, shareWith, sshKeys, deleteOn);
    }

    private Instruction generateCreateSlice() throws IOException {
        askNewSliceQuestions();
        return new CreateSliceInstruction(Instruction.Action.CREATESLICE, user, actionOutputs, debugOutputs, callOutputs, proxy, slice);
    }

    private Instruction generateListResources() throws IOException {
        out.println("\n*** ListResources questions:\n");

        boolean available = askBoolChoice("Request only available resources", "Request all resources", false);

        String testbedId = askString(
                "Target Testbed (Give either a fedmon testbed ID, a fedmon server ID, or a testbed URN)",
                null, false);

        return new ListResourcesInstruction(Instruction.Action.LISTRESOURCES, user, actionOutputs, debugOutputs, callOutputs, proxy,
                available, testbedId);
    }

    private Instruction generateManifest() throws IOException {
        askExistingSliceQuestions();
        return new ManifestInstruction(Instruction.Action.MANIFEST, user, actionOutputs, debugOutputs, callOutputs, proxy, slice);
    }

    private Instruction generateDelete() throws IOException {
        askExistingSliceQuestions();
        return new DeleteInstruction(Instruction.Action.DELETE, user, actionOutputs, debugOutputs, callOutputs, proxy, slice);
    }

    private Instruction generatePoa() throws IOException {
        askExistingSliceQuestions();

        String poaActionName = askListValue("Choose Operational Action to Perform",
                Stream.of(PoaInstruction.PoaAction.values()).map(Enum::name).collect(Collectors.toList()),
                0);
        PoaInstruction.PoaAction poaAction = PoaInstruction.PoaAction.valueOf(poaActionName);

        String sliverUrn = askString(
                "Sliver URN (you can leave this blank, which means \"all slivers\")", null, true);

        return new PoaInstruction(Instruction.Action.POA, user, actionOutputs, debugOutputs, callOutputs, proxy,
                slice,
                poaAction,
                sliverUrn);
    }

    private Instruction generateRenew() throws IOException {
        askExistingSliceQuestions(true);
        return new RenewInstruction(Instruction.Action.RENEW, user, actionOutputs, debugOutputs, callOutputs, proxy, slice);
    }

    private Instruction generateSliceCredential() throws IOException {
        askExistingSliceQuestions();
        return new SliceCredentialInstruction(Instruction.Action.SLICECREDENTIAL, user, actionOutputs, debugOutputs, callOutputs, proxy, slice);
    }

    private Instruction generateLogin() throws IOException {
        askExistingSliceQuestions();

        Boolean showCommand = askBoolChoice("Show SSH Command", "Login to the node using SSH", false);
        String clientId = askString("clientId of node to login to (leave empty to login to the only node of single nod experiments)", null, true);

        return new SshInstruction(Instruction.Action.LOGIN, user, actionOutputs, debugOutputs, callOutputs, proxy, slice,
                showCommand,
                clientId);
    }

    private Instruction generateDecryptPem() throws IOException {
        return new DecryptPemInstruction(Instruction.Action.DECRYPTPEM, user, actionOutputs, debugOutputs, callOutputs, proxy);
    }
}
